//
// Super slim Smile decoder
// Copyright(C)'2018- twaydcc / Desktop Station / MECY
//

#include <Arduino.h>
#include <string.h>
#include <avr/pgmspace.h> 
#include <avr/eeprom.h>	 //required by notifyCVRead() function if enabled below
//#include <avr/wdt.h>
#include "motor_func.h"
#include "motor_ctrl.h"
#include "NmraDcc_v142.h"


/*************************************************
 * 
 * Declaration of Symbols
 * 
 *************************************************/

//#define _DEBUG		//リリースのときはコメントアウトすること
//#define DSFOX			//DSfoxでデバッグする

#define MAN_VER_NUMBER  0   /* Release Ver CV07 */


#define DECODER_ADDRESS 3

#ifdef DSFOX
 #define PIN_LIGHT	8
 #define PIN_FHEAD	A3
 #define PIN_RHEAD	A4
#else
 #define PIN_FHEAD	6
 #define PIN_RHEAD	7
#endif

#define MOTOR_LIM_MAX 255
#define MAX_BRIGHT    3

#define CV_VSTART 2
#define CV_ACCRATIO 3
#define CV_DECCRATIO  4
#define CV_VMAX 5
#define CV_BRIGHT     33
#define CV_LIGHT      34
//#define CV_WDT  35

#define CV_BEMFcoefficient  54
#define CV_PI_P 55
#define CV_PI_I 56
#define CV_BEMFCUTOFF 57


/*************************************************
 * 
 * Declaration of Classes
 * 
 *************************************************/

NmraDcc	 Dcc;
DCC_MSG	 Packet;

/*************************************************
 * 
 * Declaration of Variables
 * 
 *************************************************/


//Task Schedule
unsigned long gPreviousL1 = 0;
//unsigned long gPreviousL2 = 0; // VVVF control
unsigned long gPreviousL2 = 0;
unsigned long gPreviousL3 = 0;
unsigned long gPreviousL4 = 0;
//unsigned long gPreviousL5 = 0; // gIntervalRail

boolean gbMyActive = false; // unpacket wdt

//モータ制御関連の変数
uint16_t gSpeedCmd = 0;
uint8_t gPwmDirv = 128;
uint8_t gCV29_Vstart = 0;
uint8_t gCV29Direction = 0;

// CV
uint8_t gCV1_SAddr = 3;
uint8_t gCVx_LAddr = 3;
uint8_t gCV2_Vstart = 20;
uint8_t gCV3_AccRatio = 16;
uint8_t gCV4_DecRatio = 16;
uint8_t gCV5_VMAX = 255;
uint8_t gCV33_Bright = 0;
uint8_t gCV34_BiDir = 1;
uint8_t gCV35_Wdtime = 0;

uint8_t gBrightCount = 0;

//ファンクションの変数
uint8_t gFuncBits[2] = {0};

struct CVPair{
  uint16_t	CV;
  uint8_t	Value;
};

CVPair FactoryDefaultCVs [] = {
  {CV_MULTIFUNCTION_PRIMARY_ADDRESS, DECODER_ADDRESS},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},		//The LSB is set CV 1 in the libraries .h file, which is the regular address location, so by setting the MSB to 0 we tell the library to use the same address as the primary address. 0 DECODER_ADDRESS
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB, 0},	 //XX in the XXYY address
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB, 0},	 //YY in the XXYY address
  {CV_29_CONFIG, 2},	 //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
//  {CV_29_CONFIG, 128},   //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
  {CV_VSTART, 20},
  {CV_ACCRATIO, 16},
  {CV_DECCRATIO, 8},
  {CV_BRIGHT, 0},           // headlight brightness 0:100%,1:75%,2:50%,3:25%
  {CV_LIGHT, 1},            // headlight ctrl. 1:bi-direction, 0:single
#ifdef CV_WDT
  {CV_WDT, 0},              // 4=WDTO_250MS, 5=WDTO_500MS, 6=WDTO_1S, 7=WDTO_2S
#endif
  {CV_BEMFcoefficient, 64}, // 0-127 step / Reajust by MECY 2017/04/16
  {CV_PI_P, 18},
  {CV_PI_I, 96},
  {CV_BEMFCUTOFF,0},        // 0でBEMF Off 1でBEMF On
};

/*************************************************
 * 
 * Declaration of Functions
 * 
 *************************************************/

void(* resetFunc) (void) = 0;  //declare reset function at address 0

uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);

void ControlRoomLight(void);
void ControlHeadLight(void);

/*************************************************
 * 
 * Functions
 * 
 *************************************************/

void notifyCVResetFactoryDefault()
{
	//When anything is writen to CV8 reset to defaults. 

	resetCVToDefault();	 
#ifdef _DEBUG
	Serial.println("Resetting...");
#endif
	delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying
	resetFunc();
}


void resetCVToDefault()
{
	//CVをデフォルトにリセット
#ifdef _DEBUG
	Serial.println("CVs being reset to factory defaults");
#endif
	
	for (int j=0; j < FactoryDefaultCVIndex; j++ ){
		Dcc.setCV( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
	}
}



extern void	   notifyCVChange( uint16_t CV, uint8_t Value)
{
   //CVが変更されたときのメッセージ
#ifdef _DEBUG
   Serial.print("CV "); 
   Serial.print(CV); 
   Serial.print(" Changed to "); 
   Serial.println(Value, DEC);
#endif
}

void notifyCVAck(void)
{
#ifdef _DEBUG
  Serial.println("notifyCVAck");
#endif
  
	MOTOR_Ack();
}


void setup()
{
	pinMode(MOTOR_PWM_A, OUTPUT);
	digitalWrite(MOTOR_PWM_A, 0);
	pinMode(MOTOR_PWM_B, OUTPUT);
	digitalWrite(MOTOR_PWM_B, 0);
	
#ifdef _DEBUG
	//シリアル通信開始
	Serial.begin(9600);
#endif

	//ファンクションの割り当てピン初期化
	#ifdef DSFOX
	pinMode(PIN_LIGHT, OUTPUT);
	digitalWrite(PIN_LIGHT, 0);
	#endif
	
	pinMode(PIN_FHEAD, OUTPUT);
	digitalWrite(PIN_FHEAD, 0);
	
	pinMode(PIN_RHEAD, OUTPUT);
	digitalWrite(PIN_RHEAD, 0);
	
	if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF )
	{
		//if eeprom has 0xFF then assume it needs to be programmed
#ifdef _DEBUG
	  Serial.println("CV Defaulting due to blank eeprom");
#endif
	  
	  notifyCVResetFactoryDefault();
	  
   } else{
#ifdef _DEBUG
	 Serial.println("CV Not Defaulting");
#endif
   }
  
	// Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
	Dcc.pin(0, 11, 0);

	// Call the main DCC Init function to enable the DCC Receiver
	Dcc.init( MAN_ID_DIY, MAN_VER_NUMBER,   FLAGS_MY_ADDRESS_ONLY , 0 ); 

	//Reset task
	gPreviousL1 = millis();
	gPreviousL2 = millis();
	gPreviousL3 = millis();
	gPreviousL4 = millis();

	//Init CVs
	gCV1_SAddr = Dcc.getCV( CV_MULTIFUNCTION_PRIMARY_ADDRESS ) ;
	gCV2_Vstart = Dcc.getCV( CV_VSTART ) ;
	gCV3_AccRatio = Dcc.getCV( CV_ACCRATIO ) ;
	gCV4_DecRatio = Dcc.getCV( CV_DECCRATIO ) ;
	gCV5_VMAX = Dcc.getCV( CV_VMAX ) ;
	gCVx_LAddr = (Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8) + Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB );
	
	gCV29_Vstart = Dcc.getCV( CV_29_CONFIG ) ;
	//cv29 Direction Check
	if ( (gCV29_Vstart & 0x01) > 0)
	{
		gCV29Direction = 1;//REVをFWDにする
	}
	else
	{
		gCV29Direction = 0;//FWDをFWDにする
	}
	
	gCV33_Bright = Dcc.getCV(CV_BRIGHT);
	gCV34_BiDir = Dcc.getCV(CV_LIGHT);
#ifdef CV_WDT
	gCV35_Wdtime = Dcc.getCV(CV_WDT);
#endif
	MOTOR_Init();
	MOTOR_SetCV(2, Dcc.getCV(CV_VSTART));
	MOTOR_SetCV(3, Dcc.getCV(CV_ACCRATIO));
	MOTOR_SetCV(4, Dcc.getCV(CV_DECCRATIO));
	MOTOR_SetCV(5, Dcc.getCV(CV_VMAX));
	MOTOR_SetCV(54, Dcc.getCV(CV_BEMFcoefficient));	//Nagoden Adjust1 2016/06/02
	MOTOR_SetCV(55, Dcc.getCV(CV_PI_P));			// MECY Adjust1 2016/04/09
	MOTOR_SetCV(56, Dcc.getCV(CV_PI_I));			// MECY Adjust1 2016/04/09
	MOTOR_SetCV(57, Dcc.getCV(CV_BEMFCUTOFF));

#ifdef _DEBUG
	Serial.print("CV1(ShortAddr): ");
	Serial.println(gCV1_SAddr);
	Serial.print("CV17/18(LongAddr): ");
	Serial.println(gCVx_LAddr);
	Serial.print("CV29: ");
	Serial.println(gCV29_Vstart);
	
	Serial.println("Ready");
#endif

}

void loop()
{
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	
	if( (millis() - gPreviousL1) >= 1)
	{
		ControlHeadLight();
		
		gPreviousL1 = millis();
	}

	if( (millis() - gPreviousL2) >= 10)
	{
		//Speed detection
		// for debug
		Serial.print(gSpeedCmd, DEC);
		Serial.print(",");
		
		MOTOR_Sensor();
		
		//Reset task
		gPreviousL2 = millis();

	}
	
	if( (millis() - gPreviousL3) >= 100)
	{
		//Motor drive control
		MOTOR_Main(gSpeedCmd, gPwmDirv);
		
		//Reset task
		gPreviousL3 = millis();
	}
	
	
	if( (millis() - gPreviousL4) >= 250)
	{	
		ControlRoomLight();
		//ControlHeadLight();
		
		//Reset task
		gPreviousL4 = millis();
	}
}


//DCC速度信号の受信によるイベント
extern void notifyDccSpeed( uint16_t Addr, DCC_ADDR_TYPE AddrType, uint8_t Speed, DCC_DIRECTION Dir, DCC_SPEED_STEPS SpeedSteps )
{
	uint16_t aSpeedRef = 0;
	
	#ifdef CV_WDT
	if(!gbMyActive)
	{
		if(gCV35_Wdtime > 0)
		{
			wdt_enable(gCV35_Wdtime);
		}
		gbMyActive = true;
	}
	
	if(gCV35_Wdtime > 0)
	{
		wdt_reset();
	}
	#endif
	
	//速度値の正規化(255を100%とする処理)
	if( Speed >= 1)
	{
		aSpeedRef = ((Speed - 1) * 255) / SpeedSteps;
	}
	else
	{
		//緊急停止信号受信時の処理 //Nagoden comment 2016/06/11
#ifdef _DEBUG
		Serial.println("***** Emagency STOP **** ");
#endif
		aSpeedRef = 0;
	}

    aSpeedRef = aSpeedRef * gCV5_VMAX / 255; 
	
	gSpeedCmd = aSpeedRef;
	
  gPwmDirv = Dir;
  if ( gCV29Direction > 0)
  {
     if ( Dir == 0)
     {
          gPwmDirv = 0;
     }
    else
    {
          gPwmDirv = 128;     
    }     
  }
  else
  {
     if ( Dir == 0)
     {
          gPwmDirv = 128;
     }
    else
    {
         gPwmDirv = 0;     
    }     
  }
  
}

//ファンクション信号受信のイベント
extern void notifyDccFunc( uint16_t Addr, DCC_ADDR_TYPE AddrType, FN_GROUP FuncGrp, uint8_t FuncState)
{
	#ifdef CV_WDT
	if(gCV35_Wdtime > 0)
	{
		wdt_reset();
	}
	#endif
	
	switch(FuncGrp)
	{
    case FN_0_4:
//      Function F00
         if  ((FuncState & FN_BIT_00) == 16)
         {
            if (gFuncBits[0] == 0)
            {
              gFuncBits[0] = 1;
            }
         }    
         else
         {
            if (gFuncBits[0] == 1)
            {
              gFuncBits[0] = 0;
            }
         }
             
//      Function F01
         if  ((FuncState & FN_BIT_01) == 1)
         {
            if (gFuncBits[1] == 0)
            {
              gFuncBits[1] = 1;
            }
         }    
         else
         {
            if (gFuncBits[1] == 1)
            {
              gFuncBits[1] = 0;
            }
         }
		break;
	}
}

void exec_function (int function, int pin, int FuncState)
{
	//digitalWrite (pin, FuncState);
}

void ControlRoomLight(void)
{
#ifdef DSFOX
	digitalWrite(PIN_LIGHT, gFuncBits[1]);
#endif
}

void ControlHeadLight(void)
{
	gBrightCount++;
	if (gBrightCount > MAX_BRIGHT)
	{
		gBrightCount = 0;
	}
	
	if(gFuncBits[0] == 0)
	{
		digitalWrite(PIN_FHEAD, 0);
		digitalWrite(PIN_RHEAD, 0);
		return;
	}
	
	if( gPwmDirv == 0)
	{
		//FWD
		digitalWrite(PIN_RHEAD, 0);
		if (gSpeedCmd <= 1)
		{
			if (gBrightCount < gCV33_Bright) {
				digitalWrite(PIN_FHEAD, 0);
			} else {
				digitalWrite(PIN_FHEAD, 1);
			}
		}
		else
		{
			digitalWrite(PIN_FHEAD, 1);
		}
	}
	else
	{
		//REV
		digitalWrite(PIN_FHEAD, 0);
		if(gCV34_BiDir == 1)
		{
			if (gSpeedCmd <= 1)
			{
				if (gBrightCount < gCV33_Bright) {
					digitalWrite(PIN_RHEAD, 0);
				} else {
					digitalWrite(PIN_RHEAD, 1);
				}
			}
			else
			{
				digitalWrite(PIN_RHEAD, 1);
			}
		}
		else
		{
			digitalWrite(PIN_RHEAD, 0);
		}
	}
}
